from __future__ import annotations

"""

Implements the MCTS + Self-Refine algorithm from
`Accessing GPT-4 level Mathematical Olympiad Solutions via Monte
Carlo Tree Self-refine with LLaMa-3 8B: A Technical Report`
by Zhang et. al.

The authors' [repo](https://github.com/trotsky1997/MathBlackBox) uses critiques,
refinements, and parent nodes' answers as conversation history.
I haven't tried it yet.

"""

"""

### MCTS任务构建

MCT 自优化算法代表了蒙特卡洛树搜索（MCTS）与大型语言模型的结合，将不同的场景执行任务过程抽象为搜索树结构。树上的节点代表当前不同视角下的选择策略，而边表示主体对自身的反思。该算法的操作流程遵循 MCTS 算法的一般模式。

具体来说，我们采用模型的先验知识，来让主体通过一系列的自身反思探索来获取自身对问题的最优答案；这种方式依赖模型的对齐偏好，我们在每种偏好上设计了一个工程框架，来完成自我对不同答案的奖励进行采样策略


1、对问题生成的子任务，生成一个合理的规划的节点
2、对每个节点创建一个MCTS任务，
3、输入 problem（总问题的子任务相关的子问题）
4、评分代码重构，将片段摘录器模块集成到一个关于_evaluate_answer逻辑提示模板，模板主要作用：将每个子问题相关的loader_cosplay_builder构建一个关于evaluate_system_prompt 的生成策略，具体的为编写一个关于带有评估的评估器，由loader_cosplay_builder方法返回场景执行器（CodeGeneratorBuilder），使用add_generator添加一个问答策略(CodeGenerator)中构成问答交互，build_executor后执行  executor.chat_run() 返回_ai_message

5、自我反思代码重构,将片段摘录器模块集成到一个关于self_refine逻辑提示模板，模板主要作用：将每个子问题相关的loader_cosplay_builder构建一个关于critic_system_prompt和refine_system_prompt的生成策略，critic_system_prompt为生成一个关于子问题相关的loader_cosplay_builder中自身不完美的评价内容，refine_system_prompt为不完美评价的思考过程和评分值。
具体的为编写一个关于带有评价的生成器和反思生成器，它们由loader_cosplay_builder方法返回场景执行器（CodeGeneratorBuilder），使用add_generator添加一个问答策略(CodeGenerator)中构成问答交互，build_executor后执行  executor.chat_run() 返回_ai_message



#### MCTS执行中的重要环节
MCTS中的约束规则如下，需要保证这些节点必须符合下面所定义的基本规则

提示约束：模型在奖励评分期间必须遵守最严格的标准。生成结果需要为JSON Response format
{
    "thought": "The thought process behind the answer.",
    "answer": "A float representing the answer to the problem."
}


高分抑制：评分节点中不存在满分反馈机制；任何超过 95 分的奖励都会按固定金额减少，以遏制过高分数。

重复采样：每次访问搜索树节点都涉及对节点奖励的重复采样，以增强自我评估的可靠性。需要注意的是，当对节点的子节点进行奖励采样时，我们也会对其父节点进行奖励采样，以增加奖励采样的样本量。

"""

import logging
import math
import os
import random
import re
import threading
from collections import deque
from enum import Enum
from typing import Tuple

import numpy as np
import tqdm
from langchain.prompts import PromptTemplate
from langchain.schema import AIMessage
from langchain_core.language_models import LanguageModelInput
from langchain_core.messages import (
    BaseMessage,
)
from langchain_core.runnables import Runnable
from pydantic import BaseModel

from dreamsboard.common.callback import call_func
from dreamsboard.common.try_parse_json_object import try_parse_json_object
from dreamsboard.document_loaders.kor_loader import KorLoader
from dreamsboard.document_loaders.protocol.ner_protocol import TaskStepRefineNode
from dreamsboard.document_loaders.structured_storyboard_loader import (
    LinkedListNode,
    StructuredStoryboard,
)
from dreamsboard.dreams.task_step_md.prompts import (
    TASK_MD_TEMPLATE,
    TASK_STEP_MD_DESC_TEMPLATE,
    TASK_STEP_MD_LIST_TEMPLATE,
    TASK_STEP_MD_TEMPLATE,
    TASK_STEP_MD_TITLE_TEMPLATE,
)
from dreamsboard.engine.generate.code_generate import (
    BaseProgramGenerator,
    QueryProgramGenerator,
)
from dreamsboard.engine.memory.mctsr.prompt import (
    RefineResponse,
    gpt_prompt_config,
)
from dreamsboard.engine.storage.storage_context import StorageContext
from dreamsboard.engine.storage.task_step_store.types import BaseTaskStepStore
from dreamsboard.engine.task_engine_builder.core import CodeGeneratorBuilder

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)

logger.addHandler(handler)


PATTERN = re.compile(r"thought:\s*(.*?)\nanswer:\s*([\d.]+)", re.DOTALL)

PLAINTEXT_PATTERN = re.compile(
    r"```plaintext?([\s\S]*?)```[\s\S]*?answer: (\d+(\.\d+)?)", re.DOTALL
)

_PROMPT_TEMPLATE_1 = PromptTemplate(
    input_variables=[
        "task_step_name",
        "task_step_level",
        "task_step_id",
        "task_step_question_answer",
    ],
    template=TASK_STEP_MD_TITLE_TEMPLATE,
)
_PROMPT_TEMPLATE_1_1 = PromptTemplate(
    input_variables=[
        "task_step_name",
        "task_step_level",
        "task_step_id",
        "task_step_question_answer",
    ],
    template=TASK_STEP_MD_DESC_TEMPLATE,
)
_PROMPT_TEMPLATE_1_2 = PromptTemplate(
    input_variables=[
        "task_step_name",
        "task_step_level",
        "task_step_id",
        "task_step_question_answer",
    ],
    template=TASK_STEP_MD_LIST_TEMPLATE,
)
_PROMPT_TEMPLATE_1_3 = PromptTemplate(
    input_variables=[
        "task_step_name",
        "task_step_level",
        "task_step_id",
        "task_step_question_answer",
    ],
    template=TASK_STEP_MD_TEMPLATE,
)
_PROMPT_TEMPLATE_2 = PromptTemplate(
    input_variables=[
        "start_task_context",
        "aemo_representation_context",
        "context_placeholder",
    ],
    template=TASK_MD_TEMPLATE,
)
ROOT_UCT_SCORE = 10_000


class MCTSNode(BaseModel):
    task_step_id: str
    answer: str
    """当前任务的会话信息"""
    parent: MCTSNode | None = None
    children: list[MCTSNode] = []
    visits: int = 0
    Q: float = 0
    reward_samples: list[int] = []

    class Config:
        arbitrary_types_allowed = True

    def add_child(self, child_node: MCTSNode):
        self.children.append(child_node)

    def __repr__(self):
        return f"MCTSNode(answer={self.answer}, Q={self.Q:.2f}, visits={self.visits})"

    def add_reward(self, reward: int):
        self.reward_samples.append(reward)
        avg_reward = np.mean(self.reward_samples)
        min_reward = np.min(self.reward_samples)

        # Average worst-case and average outcomes
        self.Q = (min_reward + avg_reward) / 2


class SelectionPolicy(Enum):
    GREEDY = 1
    IMPORTANCE_SAMPLING = 2
    PAIRWISE_IMPORTANCE_SAMPLING = 3


class MCTSr(BaseModel):
    base_path: str
    task_step_id: str
    """当前任务的id"""
    storage_context: StorageContext
    """当前任务的会话存储"""
    llm_runable: Runnable[LanguageModelInput, BaseMessage]
    kor_dreams_task_step_llm: Runnable[LanguageModelInput, BaseMessage]
    problem: str
    max_rollouts: int
    exploration_constant: float = 1.0
    max_children: int = 2
    epsilon: float = 1e-10
    reward_limit: int = 95
    excess_reward_penalty: int = 5
    selection_policy: SelectionPolicy = SelectionPolicy.GREEDY

    root: MCTSNode | None = None

    # Logs
    critiques: list[str] = []
    refinements: list[str] = []
    rewards: list[float] = []
    selected_nodes: list[MCTSNode] = []

    class Config:
        arbitrary_types_allowed = True
 
    def self_refine(self, node: MCTSNode) -> Tuple[MCTSNode, RefineResponse]:
        raise NotImplementedError()

    def _evaluate_answer(self, node: MCTSNode) -> int:
        raise NotImplementedError()

    def self_evaluate(self, node: MCTSNode):
        """Evaluate the quality of the answer. Sample `num_samples` times and average the results."""
        reward = self._evaluate_answer(node)

        if reward > self.reward_limit:
            reward -= self.excess_reward_penalty

        node.add_reward(reward)

    def backpropagate(self, node: MCTSNode):
        parent = node.parent
        while parent:
            if parent.children is None or len(parent.children) == 0:
                parent = parent.parent
                continue

            best_child_Q = max(child.Q for child in parent.children)
            parent.Q = (parent.Q + best_child_Q) / 2
            parent.visits += 1
            parent = parent.parent

    def uct(self, node: MCTSNode):
        if not node.parent:
            # Using an arbitrarily high UCT score for the root node.
            # helps to prioritize breadth.
            return ROOT_UCT_SCORE

        return node.Q + self.exploration_constant * math.sqrt(
            math.log(node.parent.visits + 1) / (node.visits + self.epsilon)
        )

    def is_fully_expanded(self, node: MCTSNode):
        return len(node.children) >= self.max_children or any(
            child.Q > node.Q for child in node.children
        )

    def select_node(self):
        """Select a non-fully expanded node with the highest UCT value.

        A node is fully expanded if either:
        1. It has reached the max number of children
        2. Any of its children have a Q value greater than its own
        """
        candidates: list[MCTSNode] = []
        to_consider = deque([self.root])

        while to_consider:
            current_node = to_consider.popleft()
            if not self.is_fully_expanded(current_node):
                candidates.append(current_node)
            to_consider.extend(current_node.children)

        if not candidates:
            return self.root

        if self.selection_policy == SelectionPolicy.GREEDY:
            return max(candidates, key=self.uct)
        elif self.selection_policy == SelectionPolicy.IMPORTANCE_SAMPLING:
            # Sample, weighted by UCT score
            uct_scores = [self.uct(node) for node in candidates]
            selected_pair_idx = random.choices(
                range(len(candidates)), weights=uct_scores, k=1
            )[0]
            return candidates[selected_pair_idx]
        elif self.selection_policy == SelectionPolicy.PAIRWISE_IMPORTANCE_SAMPLING:
            # Sample, weighted by the difference in UCT scores between pairs
            uct_scores = [self.uct(node) for node in candidates]
            pairs = [
                (i, j) for i in range(len(candidates)) for j in range(len(candidates))
            ]
            pair_weights = [
                max(uct_scores[i], uct_scores[j]) - min(uct_scores[i], uct_scores[j])
                for i, j in pairs
            ]
            selected_pair_idx = random.choices(
                range(len(pairs)), weights=pair_weights, k=1
            )[0]
            selected_candidate_idx = max(
                pairs[selected_pair_idx], key=lambda x: uct_scores[x]
            )
            return candidates[selected_candidate_idx]
        else:
            raise ValueError(f"Invalid selection policy: {self.selection_policy}")

    def initialize(self):
        """Generate a zero-shot answer."""
        # 构建MCTS树,
        structured_storyboard = _build_structured_storyboard(
            self.storage_context.task_step_store
        )

        linked_list_node = structured_storyboard.get_task_step_node(self.task_step_id)
        #  初始化当前节点self.root的上级节点
        mcts_tree = linked_list_to_tree(structured_storyboard.head)

        matching_node = find_matching_node_in_tree_iterative(
            mcts_tree, linked_list_node
        )
        self.root = matching_node

    def run(self):
        for _ in tqdm.tqdm(range(self.max_rollouts)):
            node = self.select_node()
            self.self_evaluate(node)
            child, refined_answer = self.self_refine(node)
            node.add_child(child)

            if refined_answer.answer_score > self.reward_limit:
                refined_answer.answer_score -= self.excess_reward_penalty

            child.add_reward(refined_answer.answer_score)
            self.backpropagate(child)

        owner = f"thread {threading.get_native_id()}, run end"
        logger.info(f"owner:{owner}")
        return self.get_best_answer()

    def get_best_answer(self):
        from collections import deque

        to_visit = deque([self.root])
        best_node = self.root

        while to_visit:
            current_node = to_visit.popleft()
            if current_node.Q > best_node.Q:
                best_node = current_node
            to_visit.extend(current_node.children)

        return best_node.answer

    def print(self):
        print_tree(self.root)


class MCTSrStoryboard(MCTSr):
    """

    resource_id=f"resource_{node.task_step_id}"
    """

    @staticmethod
    def _get_ai_message(callback, resource_id, **kwargs):
        code_gen_builder = CodeGeneratorBuilder.from_template(
            nodes=[], storage_context=kwargs.get("storage_context")
        )
        _base_render_data = {
            "system_prompt": kwargs.get("system_prompt"),
            "messages": [kwargs.get("user_prompt")],
        }
        code_gen_builder.add_generator(
            BaseProgramGenerator.from_config(
                cfg={
                    "code_file": "base_template_system.py-tpl",
                    "render_data": _base_render_data,
                }
            )
        )

        executor = code_gen_builder.build_executor(
            llm_runable=kwargs.get("llm_runable"), messages=[]
        )
        executor.execute()
        _ai_message = executor.chat_run()

        logger.info("\033[1;32m" + f"{resource_id}: {_ai_message}" + "\033[0m")
        assert executor._ai_message is not None

        callback(_ai_message)

    @staticmethod
    def _wrapper_steps_unit(
        context_linked_list_node: LinkedListNode, continue_task_step_id: str
    ):
        # 使用 TASK_STEP_MD_TEMPLATE 格式化每个任务步骤
        formatted_task_steps = []
        past_steps = []

        past_steps.append(
            f"{context_linked_list_node.task_step_level} task_id: {context_linked_list_node.task_step_id}, {context_linked_list_node.task_step_name}\n"
        )

        # 计算层级关系
        level_count = context_linked_list_node.task_step_level.count(">")

        if level_count == 0:
            # 一级，格式化为标题 #
            step_text = _PROMPT_TEMPLATE_1.format(
                task_step_name=f"### {context_linked_list_node.task_step_name}",
                task_step_level=context_linked_list_node.task_step_level,
                task_step_description=context_linked_list_node.task_step_description,
                task_step_id=context_linked_list_node.task_step_id,
                task_step_question_answer=context_linked_list_node.task_step_question_answer,
            )
            formatted_task_steps.append(step_text.strip() + "\n\n")
        elif level_count == 1:
            # 二级，格式化为标题 ##
            step_text = _PROMPT_TEMPLATE_1_1.format(
                task_step_name=f"{context_linked_list_node.task_step_name}",
                task_step_level=context_linked_list_node.task_step_level,
                task_step_description=context_linked_list_node.task_step_description,
                task_step_id=context_linked_list_node.task_step_id,
                task_step_question_answer=context_linked_list_node.task_step_question_answer,
            )
            formatted_task_steps.append(step_text.strip() + "\n\n")

        elif level_count >= 2:
            # 三级及以上，格式化为分类 -
            step_text = _PROMPT_TEMPLATE_1_2.format(
                task_step_name=f"- {context_linked_list_node.task_step_name}",
                task_step_level=context_linked_list_node.task_step_level,
                task_step_description=context_linked_list_node.task_step_description,
                task_step_id=context_linked_list_node.task_step_id,
                task_step_question_answer=context_linked_list_node.task_step_question_answer,
            )
            formatted_task_steps.append(step_text.strip() + "\n\n")

        else:
            step_text = _PROMPT_TEMPLATE_1_3.format(
                task_step_name=f"{context_linked_list_node.task_step_name}",
                task_step_level=context_linked_list_node.task_step_level,
                task_step_description=context_linked_list_node.task_step_description,
                task_step_id=context_linked_list_node.task_step_id,
                task_step_question_answer=context_linked_list_node.task_step_question_answer,
            )

            formatted_task_steps.append(step_text.strip())

        while (
            context_linked_list_node is not None
            and context_linked_list_node.next is not None
        ):
            if context_linked_list_node.task_step_id == continue_task_step_id:
                break
            context_linked_list_node = context_linked_list_node.next
            # 计算层级关系
            level_count = context_linked_list_node.task_step_level.count(">")

            past_steps.append(
                f"{context_linked_list_node.task_step_level} task_id: {context_linked_list_node.task_step_id}, {context_linked_list_node.task_step_name}\n"
            )

            if level_count == 0:
                # 一级，格式化为标题 #
                step_text = _PROMPT_TEMPLATE_1.format(
                    task_step_name=f"### {context_linked_list_node.task_step_name}",
                    task_step_level=context_linked_list_node.task_step_level,
                    task_step_description=context_linked_list_node.task_step_description,
                    task_step_id=context_linked_list_node.task_step_id,
                    task_step_question_answer=context_linked_list_node.task_step_question_answer,
                )
                formatted_task_steps.append(step_text.strip() + "\n\n")

            elif level_count == 1:
                # 二级，格式化为标题 ##
                step_text = _PROMPT_TEMPLATE_1_1.format(
                    task_step_name=f"{context_linked_list_node.task_step_name}",
                    task_step_level=context_linked_list_node.task_step_level,
                    task_step_description=context_linked_list_node.task_step_description,
                    task_step_id=context_linked_list_node.task_step_id,
                    task_step_question_answer=context_linked_list_node.task_step_question_answer,
                )
                formatted_task_steps.append(step_text.strip() + "\n\n")

            elif level_count >= 2:
                # 三级及以上，格式化为分类 -
                step_text = _PROMPT_TEMPLATE_1_2.format(
                    task_step_name=f"- {context_linked_list_node.task_step_name}",
                    task_step_level=context_linked_list_node.task_step_level,
                    task_step_description=context_linked_list_node.task_step_description,
                    task_step_id=context_linked_list_node.task_step_id,
                    task_step_question_answer=context_linked_list_node.task_step_question_answer,
                )
                formatted_task_steps.append(step_text.strip() + "\n\n")

            else:
                step_text = _PROMPT_TEMPLATE_1_3.format(
                    task_step_name=f"{context_linked_list_node.task_step_name}",
                    task_step_level=context_linked_list_node.task_step_level,
                    task_step_description=context_linked_list_node.task_step_description,
                    task_step_id=context_linked_list_node.task_step_id,
                    task_step_question_answer=context_linked_list_node.task_step_question_answer,
                )

                formatted_task_steps.append(step_text.strip())

        # 将格式化的步骤列表转换为字符串
        context_placeholder = "".join(formatted_task_steps)
        past_steps_placeholder = "".join(past_steps)
        return context_placeholder, past_steps_placeholder

    def self_refine(self, node: MCTSNode) -> Tuple[MCTSNode, RefineResponse]:
        """
        自我反思
        """

        critic_system_prompt_template = PromptTemplate(
            input_variables=[
                "problem",
                "current_answer",
                "context",
                "past_steps",
                "start_task_context",
                "aemo_representation_context",
                "task_step_name",
                "task_step_description",
                "task_step_level",
            ],
            template=os.environ.get(
                "critic_system_prompt_data", gpt_prompt_config.critic_system_prompt_data
            ),
        )

        structured_storyboard = _build_structured_storyboard(
            self.storage_context.task_step_store
        )

        context_linked_list_node = structured_storyboard.head
        past_context, past_steps = self._wrapper_steps_unit(
            context_linked_list_node, node.task_step_id
        )
        task_step_node = self.storage_context.task_step_store.get_task_step(
            self.task_step_id
        )
        user_prompt = critic_system_prompt_template.format(
            problem=self.problem,
            current_answer=node.answer,
            context=past_context,
            past_steps=past_steps,
            start_task_context=task_step_node.start_task_context,
            aemo_representation_context=task_step_node.aemo_representation_context,
            task_step_name=task_step_node.task_step_name,
            task_step_description=task_step_node.task_step_description,
            task_step_level=task_step_node.task_step_level,
        )

        owner = f"register_event thread {threading.get_native_id()}"
        logger.info(f"owner:{owner}")

        results = call_func(
            self._get_ai_message,
            resource_id=f"resource_critic_{self.task_step_id}",
            kwargs={
                "llm_runable": self.llm_runable,
                "system_prompt": gpt_prompt_config.critic_system_prompt,
                "user_prompt": user_prompt,
                "storage_context": self.storage_context,
            },
        )

        _ai_message = results[0]
        assert _ai_message.content is not None
        cleaned_text = re.sub(r'◁think▷.*?◁/think▷', '',_ai_message.content, flags=re.DOTALL)
        critique = cleaned_text
        assert critique is not None
        self.critiques.append(critique)

        refine_system_prompt_template = PromptTemplate(
            input_variables=[
                "problem",
                "current_answer",
                "critique",
                "context",
                "past_steps",
            ],
            template=os.environ.get(
                "refine_system_prompt_data", gpt_prompt_config.refine_system_prompt_data
            ),
        )
        refine_system_prompt = refine_system_prompt_template.format(
            problem=self.problem,
            current_answer=node.answer,
            critique=critique,
            context=past_context,
            past_steps=past_steps,
        )

        owner = f"register_event thread {threading.get_native_id()}"
        logger.info(f"owner:{owner}")

        results = call_func(
            self._get_ai_message,
            resource_id=f"resource_refine_{node.task_step_id}",
            kwargs={
                "llm_runable": self.llm_runable,
                "system_prompt": gpt_prompt_config.refine_system_prompt,
                "user_prompt": refine_system_prompt,
                "storage_context": self.storage_context,
            },
        )

        _refined_answer_response_message = results[0]
        assert _refined_answer_response_message.content is not None

        json_object = {}
        try:
            task_step_refine_node_list = self._kor_task_step_refine_builder(
                _refined_answer_response_message
            )
            # 将列表answer_score平均
            answer_score_list: list[float] = []
            for task_step_refine_node in task_step_refine_node_list:
                try:
                    answer_score_list.append(float(task_step_refine_node.answer_socre))
                except ValueError:
                    answer_score_list.append(0.0)
            answer_score = sum(answer_score_list) / len(answer_score_list)
            if answer_score <= 1:
                answer_score = answer_score * 100

            json_object.update(
                {
                    "thought": task_step_refine_node_list[0].thought,
                    "answer": task_step_refine_node_list[0].answer,
                    "answer_score": answer_score,
                }
            )
        except Exception as e:
            json_object = {
                "thought": "解析失败",
                "answer": node.answer,
                "answer_score": 0,
            }

        logger.info("\033[1;32m" + f"解析后的 JSON 对象: {json_object}" + "\033[0m")
        refined_answer = RefineResponse.model_validate(json_object)
        self.refinements.append(refined_answer)

        # ```thought \n{refined_answer.thought} ```\n\n ```answer_score \n{refined_answer.answer_score} ```
        return MCTSNode(
            answer=f"{refined_answer.answer}",
            task_step_id=self.task_step_id,
            parent=node,
            children=[],
            visits=0,
            Q=0,
            reward_samples=[],
        ), refined_answer

    def _evaluate_answer(self, node: MCTSNode) -> int:
        """
        评估答案
        """

        evaluate_system_prompt_template = PromptTemplate(
            input_variables=["problem", "past_steps", "context", "answer"],
            template=os.environ.get(
                "evaluate_system_prompt_data",
                gpt_prompt_config.evaluate_system_prompt_data,
            ),
        )
        structured_storyboard = _build_structured_storyboard(
            self.storage_context.task_step_store
        )

        context_linked_list_node = structured_storyboard.head
        past_context, past_steps = self._wrapper_steps_unit(
            context_linked_list_node, node.task_step_id
        )

        user_prompt = evaluate_system_prompt_template.format(
            problem=self.problem,
            answer=node.answer,
            context=past_context,
            past_steps=past_steps,
        )

        for attempt in range(3):
            try:
                owner = f"register_event thread {threading.get_native_id()}, _evaluate_answer"
                logger.info(f"owner:{owner}")

                results = call_func(
                    self._get_ai_message,
                    resource_id=f"resource_evaluate_{node.task_step_id}",
                    kwargs={
                        "llm_runable": self.llm_runable if self.kor_dreams_task_step_llm is None else self.kor_dreams_task_step_llm,
                        "system_prompt": gpt_prompt_config.evaluate_system_prompt,
                        "user_prompt": user_prompt,
                        "storage_context": self.storage_context,
                    },
                )

                _ai_message = results[0]

                owner = (
                    f"event end thread {threading.get_native_id()}, _evaluate_answer"
                )
                logger.info(f"owner:{owner}")
                assert _ai_message.content is not None
                cleaned_text = re.sub(r'◁think▷.*?◁/think▷', '',_ai_message.content, flags=re.DOTALL)
                return int(cleaned_text)
            except ValueError:
                user_prompt = (
                    f"{_ai_message.content}\n\nFailed to parse reward as an integer."
                )

                if attempt == 2:
                    raise

    def _kor_task_step_refine_builder(
        self, refined_answer_response_message: AIMessage
    ) -> list[TaskStepRefineNode]:
        """
        抽取根据批评意见优化当前回答并续写上下文内容
        """
        kor_task_step_refine_builder = KorLoader.form_kor_task_step_refine_builder(
            llm_runable=self.llm_runable if self.kor_dreams_task_step_llm is None else self.kor_dreams_task_step_llm,
        )
        response = kor_task_step_refine_builder.run(
            refined_answer_response_message.content
        )
        task_step_refine_node_list = []
        if (
            response.get("data") is not None
            and response.get("data").get("script") is not None
        ):
            step_list = response.get("data").get("script")
            for step in step_list:
                task_step_refine_node = TaskStepRefineNode(
                    thought=step.get("thought"),
                    answer=step.get("answer"),
                    answer_socre=step.get("answer_socre"),
                )
                task_step_refine_node_list.append(task_step_refine_node)

        return task_step_refine_node_list


def print_tree(node: MCTSNode | None, level: int = 0):
    if node is None:
        return
    indent = " " * level * 2
    node_str = repr(node)
    for line in node_str.split("\n"):
        logger.info(indent + line)
    for child in node.children:
        print_tree(child, level + 1)


def _build_structured_storyboard(
    task_step_store: BaseTaskStepStore,
) -> StructuredStoryboard:
    task_step_all = task_step_store.task_step_all
    task_step_all_list = [val.__dict__ for val in list(task_step_all.values())]
    structured_storyboard = StructuredStoryboard(json_data=task_step_all_list)

    return structured_storyboard


# 链表转树的函数
def linked_list_to_tree(linked_list_head):
    if linked_list_head is None:
        return None

    # 递归构建树
    def build_tree(node: LinkedListNode, parent: MCTSNode = None):
        if node is None:
            return None

        # 为当前链表节点创建树节点
        mcts_node = MCTSNode(
            task_step_id=node.task_step_id,
            answer=node.task_step_question_answer,
            children=[],
            visits=0,
            Q=0,
            reward_samples=[],
            parent=parent,
        )

        # 如果有下一个节点，继续递归构建子节点
        if node.next:
            child_node = build_tree(node.next, mcts_node)
            mcts_node.children.append(child_node)

        return mcts_node

    # 从链表头开始构建树
    root = build_tree(linked_list_head)
    return root


# 使用栈来模拟深度优先搜索
def find_matching_node_in_tree_iterative(tree_node, linked_list_node):
    # 初始化栈，开始时栈中只包含根节点
    stack = [tree_node]

    while stack:
        # 获取栈顶的节点
        current_node = stack.pop()

        # 如果当前节点与链表节点相同，返回当前节点
        if current_node.answer == linked_list_node.task_step_question_answer:
            return current_node

        # 将子节点添加到栈中，按深度优先顺序处理
        for child in current_node.children:
            stack.append(child)

    return None  # 如果没有匹配的节点，返回 None



import os

import weaviate
from weaviate.classes.config import Configure, DataType, Property, VectorDistances
from weaviate.classes.init import AdditionalConfig, Timeout
from weaviate.classes.query import Filter
from weaviate.client import WeaviateAsyncClient, WeaviateClient


def init_context_connect() -> WeaviateClient:
    headers = {"X-Zhipuai-Api-Key": os.environ.get("ZHIPUAI_API_KEY")}
    client = weaviate.connect_to_local(
        headers=headers,
        additional_config=AdditionalConfig(
            timeout=Timeout(init=30, query=12000, insert=120)  # Values in seconds
        ),
    )

    print(client.is_ready())
    return client


def init_context_collections(
    client: WeaviateClient, collection_id: str
) -> tuple[str, str]:
    from datetime import datetime

    # Collection name with timestamp
    collection_name = "AtomgitPapers" + collection_id

    collection_name_context = f"{collection_name}Context"
    # client.collections.delete(collection_name)
    # client.collections.delete(collection_name_context)

    if not client.collections.exists(collection_name):
        client.collections.create(
            collection_name,
            reranker_config=Configure.Reranker.transformers(),
            generative_config=Configure.Generative.zhipuai(
                # These parameters are optional
                model="glm-4-plus",
                max_tokens=500,
                temperature=0.7,
                top_p=0.7,
            ),
            vectorizer_config=[
                # Set another named vector
                Configure.NamedVectors.text2vec_transformers(
                    name="chunk_text",
                    source_properties=["chunk_text"],
                    vector_index_config=Configure.VectorIndex.hnsw(
                        distance_metric=VectorDistances.COSINE
                    ),
                ),
            ],
            properties=[  # Define properties
                Property(name="ref_id", data_type=DataType.TEXT),
                Property(name="paper_id", data_type=DataType.TEXT),
                Property(name="paper_title", data_type=DataType.TEXT),
                Property(name="chunk_id", data_type=DataType.NUMBER),
                Property(name="chunk_text", data_type=DataType.TEXT),
                Property(name="original_filename", data_type=DataType.TEXT),
            ],
        )

    if not client.collections.exists(collection_name_context):
        client.collections.create(
            collection_name_context,
            vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
                vectorize_collection_name=False
            ),
            properties=[  # Define properties
                Property(name="refId", data_type=DataType.TEXT),
                Property(name="paperId", data_type=DataType.TEXT),
                Property(name="chunkId", data_type=DataType.NUMBER),
                Property(name="chunkText", data_type=DataType.TEXT),
            ],
        )

    return collection_name, collection_name_context


def insert_into_database(
    client: WeaviateClient,
    collection_name: str,
    union_id: str,
    properties_list: List[dict],
):
    """
    插入数据到向量数据库,检查唯一
    """

    collection = client.collections.get(collection_name)
    union_ids = [item.get(union_id) for item in properties_list]
    response = collection.query.fetch_objects(
        filters=Filter.by_property(union_id).contains_any(union_ids),
    )

    exist_ids = [o.properties[union_id] for o in response.objects]

    for properties in properties_list:
        if properties[union_id] not in exist_ids:
            collection.data.insert(properties=properties)



from __future__ import annotations

import logging
import os
from abc import ABC
from typing import Any, Dict, List

from kor.extraction.parser import KorParser
from kor.nodes import Number, Object, Text
from langchain.chains import LLMChain, SequentialChain
from langchain.chains.base import Chain
from langchain_core.language_models import LanguageModelInput
from langchain_core.messages import (
    BaseMessage,
)
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import PromptTemplate
from langchain_core.runnables import Runnable, RunnableLambda, RunnableParallel

from dreamsboard.common.csv_data import CSVEncoder
from dreamsboard.document_loaders import KorLoader
from dreamsboard.document_loaders.protocol.ner_protocol import TaskStepNode
from dreamsboard.dreams.aemo_representation_chain.prompts import (
    AEMO_REPRESENTATION_PROMPT_TEMPLATE,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)


class AEMORepresentationChain(ABC):
    aemo_representation_chain: Chain
    kor_dreams_task_step_chain: LLMChain
    kor_schema: Object

    def __init__(
        self,
        start_task_context: str,
        kor_dreams_task_step_chain: LLMChain,
        kor_schema: Object,
        aemo_representation_chain: Chain,
    ):
        self.start_task_context = start_task_context
        self.kor_dreams_task_step_chain = kor_dreams_task_step_chain
        self.kor_schema = kor_schema
        self.aemo_representation_chain = aemo_representation_chain

    @classmethod
    def from_aemo_representation_chain(
        cls,
        llm_runable: Runnable[LanguageModelInput, BaseMessage],
        start_task_context: str,
        kor_dreams_task_step_llm: Runnable[LanguageModelInput, BaseMessage] = None,
    ) -> AEMORepresentationChain:
        # 00-判断情感表征是否符合.txt
        prompt_template1 = PromptTemplate(
            input_variables=["start_task_context"],
            template=os.environ.get(
                "AEMO_REPRESENTATION_PROMPT_TEMPLATE",
                AEMO_REPRESENTATION_PROMPT_TEMPLATE,
            ),
        )

        aemo_representation_chain = prompt_template1 | llm_runable | StrOutputParser()

        def wrapper_output(_dict):
            return {
                # 中间变量全部打包输出
                "aemo_representation_context": _dict["aemo_representation_context"],
            }

        aemo_representation_chain = {
            "aemo_representation_context": aemo_representation_chain,
        } | RunnableLambda(wrapper_output)

        (
            kor_dreams_task_step_chain,
            schema,
        ) = KorLoader.form_kor_dreams_task_step_builder(
            llm_runable=llm_runable
            if kor_dreams_task_step_llm is None
            else kor_dreams_task_step_llm
        )
        return cls(
            start_task_context=start_task_context,
            kor_dreams_task_step_chain=kor_dreams_task_step_chain,
            kor_schema=schema,
            aemo_representation_chain=aemo_representation_chain,
        )

    def invoke_kor_dreams_task_step_context(
        self, aemo_representation_context: str
    ) -> List[TaskStepNode]:
        """
        对开始任务进行抽取，得到任务步骤
        :return:
        """

        response = self.kor_dreams_task_step_chain.run(aemo_representation_context)
        task_step_node_list = []
        if (
            response.get("data") is not None
            and response.get("data").get("script") is not None
        ):
            step_list = response.get("data").get("script")
            for step in step_list:
                task_step_node = TaskStepNode(
                    start_task_context=self.start_task_context,
                    aemo_representation_context=aemo_representation_context,
                    task_step_name=step.get("task_step_name"),
                    task_step_description=step.get("task_step_description"),
                    task_step_level=step.get("task_step_level"),
                )
                task_step_node_list.append(task_step_node)
        else:
            encoder = CSVEncoder(node=self.kor_schema)
            parser = KorParser(encoder=encoder, schema_=self.kor_schema)
            response = parser.parse(response.get("raw"))

            if (
                response.get("data") is not None
                and response.get("data").get("script") is not None
            ):
                step_list = response.get("data").get("script")
                for step in step_list:
                    task_step_node = TaskStepNode(
                        start_task_context=self.start_task_context,
                        aemo_representation_context=aemo_representation_context,
                        task_step_name=step.get("task_step_name"),
                        task_step_description=step.get("task_step_description"),
                        task_step_level=step.get("task_step_level"),
                    )
                    task_step_node_list.append(task_step_node)
        return task_step_node_list

    def invoke_aemo_representation_context(self) -> Dict[str, Any]:
        return self.aemo_representation_chain.invoke(
            {"start_task_context": self.start_task_context}
        )



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

from dreamsboard.engine.schema import BaseNode, ObjectTemplateType
from dreamsboard.templates import get_template_path

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)


class TaskStepContext(BaseModel):
    ref_id: str = Field(default="", description="ref_id")
    paper_title: str = Field(default="", description="paper_title")
    chunk_id: str = Field(default="", description="chunk_id")
    score: float = Field(default=0.0, description="score")
    text: str = Field(default="", description="text")


class TaskStepNode(BaseNode, ABC):
    """任务步骤节点"""

    start_task_context: Optional[str] = Field(default="", description="开始任务")
    aemo_representation_context: Optional[str] = Field(
        default="", description="任务总体描述"
    )
    task_step_name: Optional[str] = Field(default="", description="任务步骤名称")
    task_step_description: Optional[str] = Field(default="", description="任务步骤描述")
    task_step_level: Optional[str] = Field(default="", description="任务步骤层级")
    task_step_question: Optional[str] = Field(default="", description="任务步骤问题")
    task_step_question_context: Optional[List[TaskStepContext]] = Field(
        default=[], description="任务步骤问题上下文"
    )

    task_step_question_answer: Optional[str] = Field(
        default="", description="任务步骤答案"
    )
    ref_task_step_id: Optional[str] = Field(
        default="",
        description="ref_task_step_id",
    )

    def __init__(
        self,
        start_task_context: str = None,
        aemo_representation_context: str = None,
        task_step_name: str = None,
        task_step_description: str = None,
        task_step_level: str = None,
        task_step_question: str = None,
        task_step_question_context: List[TaskStepContext] = None,
        task_step_question_answer: str = None,
        ref_task_step_id: str = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.start_task_context = start_task_context
        self.aemo_representation_context = aemo_representation_context
        self.task_step_name = task_step_name
        self.task_step_description = task_step_description
        self.task_step_level = task_step_level
        self.task_step_question = task_step_question
        self.task_step_question_context = task_step_question_context
        self.task_step_question_answer = task_step_question_answer
        self.ref_task_step_id = ref_task_step_id

    @classmethod
    def from_config(cls, cfg=None):
        if cfg is None:
            raise RuntimeError("from_config cfg is None.")
        start_task_context = cfg.get("start_task_context", "")
        aemo_representation_context = cfg.get("aemo_representation_context", "")
        task_step_name = cfg.get("task_step_name", "")
        task_step_description = cfg.get("task_step_description", "")
        task_step_level = cfg.get("task_step_level", "")
        task_step_question = cfg.get("task_step_question", "")
        task_step_question_context = cfg.get("task_step_question_context", [])
        task_step_question_answer = cfg.get("task_step_question_answer", "")
        ref_task_step_id = cfg.get("ref_task_step_id", "")
        return cls(
            start_task_context=start_task_context,
            aemo_representation_context=aemo_representation_context,
            task_step_name=task_step_name,
            task_step_description=task_step_description,
            task_step_level=task_step_level,
            task_step_question=task_step_question,
            task_step_question_context=task_step_question_context,
            task_step_question_answer=task_step_question_answer,
            ref_task_step_id=ref_task_step_id,
        )

    @classmethod
    def get_type(cls) -> str:
        """Get Object type."""
        return "task_step_node"

    def class_name(self) -> str:
        """Get class name."""
        if hasattr(self, "__class_getitem__"):
            return self.__class__.__name__
        elif hasattr(self, "__orig_class__"):
            return self.__orig_class__.__name__
        elif hasattr(self, "__name__"):
            return self.__name__
        else:
            return "任务步骤节点"

    @property
    def template_content(self) -> str:
        return (
            f"start_task_context: {self.start_task_context}, "
            f"aemo_representation_context: {self.aemo_representation_context}, "
            f"task_step_name: {self.task_step_name}, "
            f"task_step_description: {self.task_step_description}, "
            f"task_step_level: {self.task_step_level}, "
            f"task_step_question: {self.task_step_question}, "
            f"task_step_question_context: {self.task_step_question_context}, "
            f"task_step_question_answer: {self.task_step_question_answer}, "
            f"ref_task_step_id: {self.ref_task_step_id}"
        )

    @template_content.setter
    def template_content(self, _template_content) -> None:
        raise RuntimeError("template_content is read only.")



import logging

import langchain
from langchain.prompts import PromptTemplate
from langchain_core.prompt_values import StringPromptValue
from langchain_core.runnables import RunnableLambda, RunnableParallel

from dreamsboard.dreams.task_step_md.prompts import (
    TASK_MD_TEMPLATE,
    TASK_REF_TEMPLATE,
    TASK_STEP_MD_DESC_TEMPLATE,
    TASK_STEP_MD_LIST_TEMPLATE,
    TASK_STEP_MD_TEMPLATE,
    TASK_STEP_MD_TITLE_TEMPLATE,
)
from dreamsboard.engine.storage.task_step_store.types import BaseTaskStepStore

logger = logging.getLogger(__name__)

_PROMPT_TEMPLATE_1 = PromptTemplate(
    input_variables=[
        "task_step_name",
        "task_step_level",
        "task_step_id",
        "task_step_question_answer",
    ],
    template=TASK_STEP_MD_TITLE_TEMPLATE,
)
_PROMPT_TEMPLATE_1_1 = PromptTemplate(
    input_variables=[
        "task_step_name",
        "task_step_level",
        "task_step_id",
        "task_step_question_answer",
    ],
    template=TASK_STEP_MD_DESC_TEMPLATE,
)
_PROMPT_TEMPLATE_1_2 = PromptTemplate(
    input_variables=[
        "task_step_name",
        "task_step_level",
        "task_step_id",
        "task_step_question_answer",
    ],
    template=TASK_STEP_MD_LIST_TEMPLATE,
)
_PROMPT_TEMPLATE_1_3 = PromptTemplate(
    input_variables=[
        "task_step_name",
        "task_step_level",
        "task_step_id",
        "task_step_question_answer",
    ],
    template=TASK_STEP_MD_TEMPLATE,
)
_PROMPT_TEMPLATE_2 = PromptTemplate(
    input_variables=[
        "start_task_context",
        "aemo_representation_context",
        "context_placeholder",
    ],
    template=TASK_MD_TEMPLATE,
)

_REF_TEMPLATE = PromptTemplate(
    input_variables=["task_step_level", "ref_id",  "paper_title", "chunk_id", "score"],
    template=TASK_REF_TEMPLATE,
)


class TaskStepMD:
    def __init__(self, task_step_store: BaseTaskStepStore):
        self.task_step_store = task_step_store

    def format_md(self) -> StringPromptValue:
        def wrapper_steps_unit(dict_input: dict):
            # 使用 TASK_STEP_MD_TEMPLATE 格式化每个任务步骤
            formatted_task_steps = []
            ref_list = []
            for step in list(self.task_step_store.task_step_all.values()):
                # 计算层级关系
                level_count = step.task_step_level.count(">")
                step_ref = []
                for ref_data in step.task_step_question_context:
                    ref_txt = _REF_TEMPLATE.format(
                        task_step_level=step.task_step_level,
                        ref_id=ref_data["ref_id"],
                        chunk_id=ref_data["chunk_id"],
                        paper_title=ref_data.get("paper_title"),
                        score=ref_data["score"],
                    )
                    ref_list.append(ref_txt)
  
                if level_count == 0:
                    # 一级，格式化为标题 #
                    step_text = _PROMPT_TEMPLATE_1.format(
                        task_step_name=f"### {step.task_step_name}",
                        task_step_level=step.task_step_level,
                        task_step_description=step.task_step_description,
                        task_step_id=step.node_id,
                        task_step_question_answer=step.task_step_question_answer,
                    )
                    formatted_task_steps.append(step_text.strip() + "\n\n")

                elif level_count == 1:
                    # 二级，格式化为标题 ##
                    step_text = _PROMPT_TEMPLATE_1_1.format(
                        task_step_name=f"{step.task_step_name}",
                        task_step_level=step.task_step_level,
                        task_step_description=step.task_step_description,
                        task_step_id=step.node_id,
                        task_step_question_answer=step.task_step_question_answer,
                    )
                    formatted_task_steps.append(step_text.strip() + "\n\n")

                elif level_count >= 2:
                    # 三级及以上，格式化为分类 -
                    step_text = _PROMPT_TEMPLATE_1_2.format(
                        task_step_name=f"- {step.task_step_name}",
                        task_step_level=step.task_step_level,
                        task_step_description=step.task_step_description,
                        task_step_id=step.node_id,
                        task_step_question_answer=step.task_step_question_answer,
                    )
                    formatted_task_steps.append(step_text.strip() + "\n\n")

                else:
                    step_text = _PROMPT_TEMPLATE_1_3.format(
                        task_step_name=f"{step.task_step_name}",
                        task_step_level=step.task_step_level,
                        task_step_description=step.task_step_description,
                        task_step_id=step.node_id,
                        task_step_question_answer=step.task_step_question_answer,
                    )

                    formatted_task_steps.append(step_text.strip())

            # 将格式化的步骤列表转换为字符串
            context_placeholder = "".join(formatted_task_steps)
            references = "\r\n".join(ref_list)
            return {
                "start_task_context": list(self.task_step_store.task_step_all.values())[
                    0
                ].start_task_context,
                "aemo_representation_context": list(
                    self.task_step_store.task_step_all.values()
                )[0].aemo_representation_context,
                "context_placeholder": context_placeholder,
                "references": references,
            }

        chain = RunnableLambda(wrapper_steps_unit) | _PROMPT_TEMPLATE_2

        out = chain.invoke({})
        return out

    def write_md(self, output_path: str) -> StringPromptValue:
        md = self.format_md()
        with open(output_path, "w", encoding="utf-8") as f:
            f.write(md.text)
        logger.info(f"Write MD to {output_path}")
        return md



from __future__ import annotations

import logging
import os
import re
import threading
from abc import ABC
from typing import Dict, List

import pandas as pd
from langchain.chains import SequentialChain
from langchain.chains.base import Chain
from langchain_core.language_models import LanguageModelInput
from langchain_core.messages import (
    BaseMessage,
)
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import PromptTemplate
from langchain_core.runnables import Runnable, RunnableLambda, RunnableParallel
from sentence_transformers import CrossEncoder

from dreamsboard.common.callback import event_manager
from dreamsboard.document_loaders import KorLoader
from dreamsboard.document_loaders.protocol.ner_protocol import TaskStepNode
from dreamsboard.dreams.task_step_to_question_chain.prompts import (
    CONVERT_TASK_STEP_TO_QUESTION_PROMPT_TEMPLATE,
    TASK_STEP_QUESTION_TO_GRAPHQL_PROMPT_TEMPLATE,
)
from dreamsboard.dreams.task_step_to_question_chain.searx.searx import searx_query
from dreamsboard.dreams.task_step_to_question_chain.weaviate.prepare_load import (
    exe_query,
    get_query_hash,
)
from dreamsboard.engine.entity.task_step.task_step import TaskStepContext
from dreamsboard.engine.storage.storage_context import BaseTaskStepStore
from dreamsboard.engine.storage.task_step_store.types import DEFAULT_PERSIST_FNAME
from dreamsboard.engine.utils import concat_dirs
from dreamsboard.vector.base import CollectionService, DocumentWithVSId

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)

"""
#### 场景加载模块

编写符合计算机科学领域的 故事情境提示词，生成研究情境（story_scenario_context），替换现有的langchain会话模板，
对每个子任务指令转换为子问题
召回问题前3条,存入task_step_question_context
调用llm，生成task_step_question_answer
"""

PATTERN = re.compile(r"```graphql?([\s\S]*?)```", re.DOTALL)
"""Regex pattern to parse the output."""


class TaskStepToQuestionChain(ABC):
    base_path: str
    start_task_context: str
    task_step_to_question_chain: Chain
    task_step_store: BaseTaskStepStore
    task_step_question_to_graphql_chain: Chain
    collection: CollectionService
    cross_encoder: CrossEncoder
    data_base: str

    def __init__(
        self,
        base_path: str,
        start_task_context: str,
        task_step_store: BaseTaskStepStore,
        task_step_to_question_chain: Chain,
        task_step_question_to_graphql_chain: Chain,
        collection: CollectionService,
        cross_encoder: CrossEncoder,
        data_base: str,
    ):
        self.base_path = base_path
        self.start_task_context = start_task_context
        self.task_step_store = task_step_store
        self.task_step_to_question_chain = task_step_to_question_chain
        self.task_step_question_to_graphql_chain = task_step_question_to_graphql_chain
        self.collection = collection
        self.cross_encoder = cross_encoder
        self.data_base = data_base

    @classmethod
    def from_task_step_to_question_chain(
        cls,
        base_path: str,
        start_task_context: str,
        llm_runable: Runnable[LanguageModelInput, BaseMessage],
        task_step_store: BaseTaskStepStore,
        collection: CollectionService,
        cross_encoder: CrossEncoder,
        data_base: str = "search_papers",
    ) -> TaskStepToQuestionChain:
        """

        1、对当前主任务名称创建hash值，作为collection_name
        2、对当前任务步骤名称创建hash值，作为collection_name_context
        """
        prompt_template1 = PromptTemplate(
            input_variables=[
                "start_task_context",
                "aemo_representation_context",
                "task_step_name",
                "task_step_description",
                "task_step_level",
            ],
            template=os.environ.get(
                "CONVERT_TASK_STEP_TO_QUESTION_PROMPT_TEMPLATE",
                CONVERT_TASK_STEP_TO_QUESTION_PROMPT_TEMPLATE,
            ),
        )

        task_step_to_question_chain = prompt_template1 | llm_runable | StrOutputParser()

        def wrapper_output(_dict):
            return {
                # 中间变量全部打包输出
                "task_step_question_context": _dict["task_step_question_context"],
            }

        task_step_to_question_chain = {
            "task_step_question_context": task_step_to_question_chain,
        } | RunnableLambda(wrapper_output)

        prompt_template2 = PromptTemplate(
            input_variables=["collection_name_context", "task_step_question"],
            template=os.environ.get(
                "TASK_STEP_QUESTION_TO_GRAPHQL_PROMPT_TEMPLATE",
                TASK_STEP_QUESTION_TO_GRAPHQL_PROMPT_TEMPLATE,
            ),
        )

        task_step_question_to_graphql_chain = (
            prompt_template2 | llm_runable | StrOutputParser()
        )

        def wrapper_output2(_dict):
            return {
                # 中间变量全部打包输出
                "task_step_question_graphql_context": _dict[
                    "task_step_question_graphql_context"
                ],
            }

        task_step_question_to_graphql_chain = {
            "task_step_question_graphql_context": task_step_question_to_graphql_chain,
        } | RunnableLambda(wrapper_output2)

        return cls(
            base_path=base_path,
            start_task_context=start_task_context,
            task_step_store=task_step_store,
            task_step_to_question_chain=task_step_to_question_chain,
            task_step_question_to_graphql_chain=task_step_question_to_graphql_chain,
            collection=collection,
            cross_encoder=cross_encoder,
            data_base=data_base,
        )

    def invoke_task_step_to_question(self, task_step_id: str) -> None:
        """
        对开始任务进行抽取，得到任务步骤
        :return:
        """

        task_step_node = self.task_step_store.get_task_step(task_step_id)
        result = self.task_step_to_question_chain.invoke(task_step_node.__dict__)
        cleaned_text = re.sub(r'◁think▷.*?◁/think▷', '', result["task_step_question_context"], flags=re.DOTALL)
        task_step_node.task_step_question = cleaned_text
        self.task_step_store.add_task_step([task_step_node])

        # 每处理一个任务步骤，就持久化一次
        task_step_store_path = concat_dirs(
            dirname=f"{self.base_path}/storage/{task_step_id}",
            basename=DEFAULT_PERSIST_FNAME,
        )
        self.task_step_store.persist(persist_path=task_step_store_path)

    @staticmethod
    def _into_database_query(callback, resource_id, **kwargs) -> None:
        """
        插入数据到向量数据库,检查唯一
        :param union_id_key:  唯一标识
        :param page_content_key:  数据列表
        :param properties_list:  数据列表
        :return: None
        """

        union_ids = [
            str(item.get(kwargs.get("union_id_key")))
            for item in kwargs.get("properties_list")
        ]

        response = kwargs.get("collection").get_doc_by_ids(ids=union_ids)

        exist_ids = [o.metadata[kwargs.get("union_id_key")] for o in response]

        docs = []
        for item in kwargs.get("properties_list"):
            metadata = {
                key: value
                for key, value in item.items()
                if key != kwargs.get("page_content_key")
            }
            if item.get(kwargs.get("union_id_key")) not in exist_ids:
                doc = DocumentWithVSId(
                    id=item.get(kwargs.get("union_id_key")),
                    page_content=item.get(kwargs.get("page_content_key")),
                    metadata=metadata,
                )
                docs.append(doc)

        kwargs.get("collection").do_add_doc(docs)

        kwargs.get("collection").save_vector_store()
        # 召回
        response = kwargs.get("collection").do_search(
            query=kwargs.get("task_step_question"), top_k=10, score_threshold=0.6
        )
        callback(response)

    def invoke_task_step_question_context(self, task_step_id: str) -> None:
        """
        对任务步骤进行抽取，得到任务步骤的上下文
        3、增加项目
        """ 

        task_step_node = self.task_step_store.get_task_step(task_step_id)
        if (
            task_step_node.task_step_question_context is not None
            and len(task_step_node.task_step_question_context) > 0
        ):
            return

        top_k = 30  # 可选参数，默认查询返回前 30个结果
        if self.data_base == "search_papers":
            properties_list = exe_query(task_step_node.task_step_question, top_k)
        elif self.data_base == "searx":
            properties_list = searx_query(task_step_node.task_step_question, top_k)

        # 插入数据到数据库
        owner = f"register_event thread {threading.get_native_id()}"
        logger.info(f"owner:{owner}")
        event_id = event_manager.register_event(
            self._into_database_query,
            resource_id=f"resource_collection_{self.collection.kb_name}",
            kwargs={
                "collection": self.collection,
                "union_id_key": "ref_id",
                "page_content_key": "chunk_text",
                "properties_list": properties_list,
                "task_step_question": task_step_node.task_step_question,
            },
        )
        results = None
        while results is None or len(results) == 0:
            results = event_manager.get_results(event_id)
        response = results[0]

        chunk_texts = []
        ref_ids = []
        chunk_ids = []
        paper_titles = []
        for o in response:
            chunk_texts.append(o.page_content)
            ref_ids.append(o.metadata["ref_id"])
            chunk_ids.append(o.metadata["chunk_id"])
            paper_titles.append(o.metadata.get("paper_title", ""))

        rankings = self.cross_encoder.rank(
            task_step_node.task_step_question,
            chunk_texts,
            show_progress_bar=True,
            return_documents=True,
            convert_to_tensor=True,
        )

        task_step_question_context = []
        # 召回问题前3条，存入task_step_question_context
        for i, ranking in enumerate(rankings[:3]):
            logger.info(
                f"ref_ids: {ref_ids[i]},chunk_ids: {chunk_ids[i]}, Score: {ranking['score']:.4f}, Text: {ranking['text']}"
            )
            task_step_question_context.append(
                TaskStepContext(
                    ref_id=str(ref_ids[i]),
                    paper_title=str(paper_titles[i]),
                    chunk_id=str(chunk_ids[i]),
                    score=ranking["score"],
                    text=ranking["text"],
                )
            )

        # rankingscorpus_id 的索引与ref_ids的索引一致
        task_step_node.task_step_question_context = task_step_question_context
        self.task_step_store.add_task_step([task_step_node])

        # 每处理一个任务步骤，就持久化一次
        task_step_store_path = concat_dirs(
            dirname=f"{self.base_path}/storage/{task_step_id}",
            basename=DEFAULT_PERSIST_FNAME,
        )
        self.task_step_store.persist(persist_path=task_step_store_path)

    def export_csv_file_path(self, task_step_id: str) -> str:
        """
        3、对召回内容与问题 导出csv文件
        """
        task_step_node = self.task_step_store.get_task_step(task_step_id)

        table_data = []
        row = [
            task_step_id,
            task_step_node.task_step_name,
            task_step_node.task_step_level,
        ]

        table_data.append(row)
        row2 = [
            task_step_id,
            task_step_node.task_step_question,
            task_step_node.task_step_level,
        ]
        table_data.append(row2)
        for context in task_step_node.task_step_question_context:
            row3 = [
                task_step_id,
                f"ref_ids: {context['ref_id']}, chunk_ids: {context['chunk_id']}, Score: {context['score']:.4f}, Text: {context['text']}",
                task_step_node.task_step_level,
            ]
            table_data.append(row3)

        table = pd.DataFrame(table_data, columns=["角色", "内容", "分镜"])

        table.to_csv(
            f"{self.base_path}/storage/{task_step_id}.csv", index=False, escapechar="\\"
        )

        return f"{self.base_path}/storage/{task_step_id}.csv"



from typing import Dict, Type

from dreamsboard.engine.generate.code_generate import (
    AIProgramGenerator,
    BaseProgramGenerator,
    CodeGenerator,
    EngineProgramGenerator,
    QueryProgramGenerator,
)
from dreamsboard.engine.schema import ObjectTemplateType

TEMPLATE_TYPE_TO_GENERATOR_CLASS: Dict[ObjectTemplateType, Type[CodeGenerator]] = {
    ObjectTemplateType.BaseProgramGenerator: BaseProgramGenerator,
    ObjectTemplateType.QueryProgramGenerator: QueryProgramGenerator,
    ObjectTemplateType.AIProgramGenerator: AIProgramGenerator,
    ObjectTemplateType.EngineProgramGenerator: EngineProgramGenerator,
}



TASK_MD_TEMPLATE = """
# {start_task_context} 


{context_placeholder}


# References  

{references}
"""


TASK_STEP_MD_TITLE_TEMPLATE = """
{task_step_name} [task_id]({task_step_id})<sup>{task_step_level}</sup>

{task_step_question_answer}
"""


TASK_STEP_MD_DESC_TEMPLATE = """
{task_step_name} [task_id]({task_step_id})<sup>{task_step_level}</sup> {task_step_question_answer}
"""


TASK_STEP_MD_LIST_TEMPLATE = """
- **{task_step_name} [task_id]({task_step_id})<sup>{task_step_level}</sup>**:{task_step_question_answer}
"""


TASK_STEP_MD_TEMPLATE = """
{task_step_name} [task_id]({task_step_id})<sup>{task_step_level}</sup>, {task_step_question_answer}
"""


TASK_REF_TEMPLATE = """[{task_step_level}] {paper_title} ,chunk_id:{ref_id} 
"""



from __future__ import annotations

import logging
import os
from abc import ABC
from typing import Any, Dict

from langchain.chains import SequentialChain
from langchain.chains.base import Chain
from langchain_core.language_models import LanguageModelInput
from langchain_core.messages import (
    BaseMessage,
)
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import PromptTemplate
from langchain_core.runnables import Runnable, RunnableLambda, RunnableParallel

from dreamsboard.document_loaders import StructuredStoryboardCSVBuilder
from dreamsboard.dreams.dreams_personality_chain.prompts import (
    DREAMS_GEN_TEMPLATE,
    EDREAMS_EVOLUTIONARY_TEMPLATE,
    EDREAMS_PERSONALITY_TEMPLATE,
    STORY_BOARD_SCENE_TEMPLATE,
    STORY_BOARD_SUMMARY_CONTEXT_TEMPLATE,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)


class StoryBoardDreamsGenerationChain(ABC):
    builder: StructuredStoryboardCSVBuilder
    dreams_guidance_chain: Chain
    dreams_personality_chain: Chain

    def __init__(
        self,
        csv_file_path: str,
        user_id: str,
        dreams_guidance_chain: Chain,
        dreams_personality_chain: Chain,
    ):
        self.builder = StructuredStoryboardCSVBuilder.form_builder(
            csv_file_path=csv_file_path
        )
        self.user_id = user_id
        self.dreams_guidance_chain = dreams_guidance_chain
        self.dreams_personality_chain = dreams_personality_chain

    @classmethod
    def from_dreams_personality_chain(
        cls,
        llm_runable: Runnable[LanguageModelInput, BaseMessage],
        csv_file_path: str,
        user_id: str = None,
    ) -> StoryBoardDreamsGenerationChain:
        # 03- 故事情境生成 `story_scenario_context`.txt STORY_BOARD_SCENE_TEMPLATE_Chain
        prompt_template1 = PromptTemplate(
            input_variables=["scene_content"], template=STORY_BOARD_SCENE_TEMPLATE
        )

        review_chain1 = prompt_template1 | llm_runable | StrOutputParser()

        # 03-故事场景生成 `scene_monologue_context`.txt
        prompt_template2 = PromptTemplate(
            input_variables=["story_board_summary_context", "user_id"],
            template=os.environ.get(
                "STORY_BOARD_SUMMARY_CONTEXT_TEMPLATE",
                STORY_BOARD_SUMMARY_CONTEXT_TEMPLATE,
            ),
        )
        review_chain2 = prompt_template2 | llm_runable | StrOutputParser()
        # 04-情感情景引导.txt
        prompt_template = PromptTemplate(
            input_variables=[
                "story_board_summary_context",
                "story_scenario_context",
                "scene_monologue_context",
                "user_id",
            ],
            template=os.environ.get("DREAMS_GEN_TEMPLATE", DREAMS_GEN_TEMPLATE),
        )
        social_chain = prompt_template | llm_runable | StrOutputParser()

        def wrapper_guidance_unit(_dict):
            return {
                # 中间变量包装至下一个pipline
                "story_scenario_context": _dict["story_scenario_context"],
                "scene_monologue_context": _dict["scene_monologue_context"],
                "user_id": _dict["user_id"],
                "scene_content": _dict["scene_content"],
                "story_board_summary_context": _dict["story_board_summary_context"],
            }

        def wrapper_guidance_output(_dict):
            return {
                # 中间变量全部打包输出
                "dreams_guidance_context": _dict["dreams_guidance_context"],
                "story_scenario_context": _dict["story_scenario_context"],
                "scene_monologue_context": _dict["scene_monologue_context"],
                "user_id": _dict["user_id"],
                "scene_content": _dict["scene_content"],
                "story_board_summary_context": _dict["story_board_summary_context"],
            }

        dreams_guidance_chain = (
            {
                "story_scenario_context": review_chain1,
                "scene_monologue_context": review_chain2,
                "user_id": lambda x: x["user_id"],
                "scene_content": lambda x: x["scene_content"],
                "story_board_summary_context": lambda x: x[
                    "story_board_summary_context"
                ],
            }
            | RunnableLambda(wrapper_guidance_unit)
            | {
                "dreams_guidance_context": social_chain,
                "story_scenario_context": lambda x: x["story_scenario_context"],
                "scene_monologue_context": lambda x: x["scene_monologue_context"],
                "user_id": lambda x: x["user_id"],
                "scene_content": lambda x: x["scene_content"],
                "story_board_summary_context": lambda x: x[
                    "story_board_summary_context"
                ],
            }
            | RunnableLambda(wrapper_guidance_output)
        )

        # 05-剧情总结 `evolutionary_step`.txt
        prompt_template04 = PromptTemplate(
            input_variables=["story_board_summary_context"],
            template=os.environ.get(
                "EDREAMS_EVOLUTIONARY_TEMPLATE", EDREAMS_EVOLUTIONARY_TEMPLATE
            ),
        )
        evolutionary_chain = prompt_template04 | llm_runable | StrOutputParser()
        # 05-性格分析 `dreams_personality_context`.txt
        prompt_template05 = PromptTemplate(
            input_variables=["evolutionary_step"],
            template=os.environ.get(
                "EDREAMS_PERSONALITY_TEMPLATE", EDREAMS_PERSONALITY_TEMPLATE
            ),
        )
        personality_chain = prompt_template05 | llm_runable | StrOutputParser()

        def wrapper_personality_unit(_dict):
            return {
                # 中间变量包装至下一个pipline
                "evolutionary_step": _dict["evolutionary_step"],
                "story_board_summary_context": _dict["story_board_summary_context"],
            }

        def wrapper_personality_output(_dict):
            return {
                # 中间变量全部打包输出
                "dreams_personality_context": _dict["dreams_personality_context"],
                "evolutionary_step": _dict["evolutionary_step"],
                "story_board_summary_context": _dict["story_board_summary_context"],
            }

        dreams_personality_chain = (
            {
                "story_board_summary_context": lambda x: x[
                    "story_board_summary_context"
                ],
                "evolutionary_step": evolutionary_chain,
            }
            | RunnableLambda(wrapper_personality_unit)
            | {
                "dreams_personality_context": personality_chain,
                "evolutionary_step": lambda x: x["evolutionary_step"],
                "story_board_summary_context": lambda x: x[
                    "story_board_summary_context"
                ],
            }
            | RunnableLambda(wrapper_personality_output)
        )

        return cls(
            csv_file_path=csv_file_path,
            user_id=user_id if user_id else "此来访者",
            dreams_guidance_chain=dreams_guidance_chain,
            dreams_personality_chain=dreams_personality_chain,
        )

    def run(self) -> Dict[str, Any]:
        # 对传入的剧本台词转换成 scene_content
        self.builder.load()
        selected_columns = ["story_board_role", "story_board_text", "story_board"]
        scene_content = self.builder.build_text(self.user_id, selected_columns)
        story_board_summary_context = self.builder.build_msg()

        dreams_guidance_personality_map_chain = RunnableParallel(
            dreams_guidance_context=self.dreams_guidance_chain,
            dreams_personality_context=self.dreams_personality_chain,
        )

        return dreams_guidance_personality_map_chain.invoke(
            {
                "scene_content": scene_content,
                "story_board_summary_context": story_board_summary_context,
                "user_id": self.user_id,
            }
        )

    def invoke_dreams_guidance_chain(self, input: Dict[str, Any]) -> Dict[str, Any]:
        return self.dreams_guidance_chain.invoke(input)

    def invoke_dreams_personality_chain(self, input: Dict[str, Any]) -> Dict[str, Any]:
        return self.dreams_personality_chain.invoke(input)
